use crate::compound::peano::{Peano, PeanoNonZero, P0, PNext};

mod flat;
// mod linked;
// mod tuple_consumer;
// pub use tuple_consumer::*;
// pub use linked::*;
pub use flat::{IntoList, Prependable, Prepended};
pub use flat::exports::*;

pub type Empty = flat::Null;

/// something which can be indexed at `P`.
/// a List of length N is expected to implement `Indexable<P>` for all `P < N`.
pub trait Indexable<P: Peano> {
    type Element;
    fn get(&self) -> Self::Element where Self::Element: Copy;
    fn get_ref(&self) -> &Self::Element;
    fn get_mut(&mut self) -> &mut Self::Element;
    fn set(&mut self, v: Self::Element);
}

/// convenience to lookup the type of the element at index `P` of list `L`.
pub type ElementAt<P, L> = <L as Indexable<P>>::Element;

/// implemented by any List.
pub trait Meta {
    type Length: PeanoNonZero;
}

/// implemented by any List, or Null (empty list)
pub trait MaybeMeta {
    type Length: Peano;
}

impl<M: Meta> MaybeMeta for M {
    type Length = M::Length;
}

/// implement on your own type for all `N` of a given list if you want to be able to walk the list.
pub trait Visitor<N: Peano, Arg> {
    fn visit(&mut self, a: &Arg);
}

/// convenience for visitors who don't care which index they're visiting; only which type.
pub trait AgnosticVisitor<Arg> {
    fn visit(&mut self, a: &Arg);
}
impl<N: Peano, Arg, V> Visitor<N, Arg> for V
where
    V: AgnosticVisitor<Arg>
{
    fn visit(&mut self, a: &Arg) {
        AgnosticVisitor::<Arg>::visit(self, a);
    }
}

/// used internally to simplify trait bounds.
/// visits every element in the list at index less than `P`.
pub trait PartialListVisitor<L, P> {
    fn visit_partial(&mut self, l: &L);
}

impl<L, V> PartialListVisitor<L, P0> for V
{
    fn visit_partial(&mut self, _unused_basecase: &L) {
    }
}

impl<L, V, P: Peano> PartialListVisitor<L, PNext<P>> for V
where
    L: Indexable<P>,
    V: PartialListVisitor<L, P> + Visitor<P, L::Element>,
{
    fn visit_partial(&mut self, l: &L) {
        PartialListVisitor::<L, P>::visit_partial(self, l);
        self.visit(l.get_ref());
    }
}

/// marker trait for a type which can visit every element in the list `L`.
pub trait ListVisitor<L: MaybeMeta>: PartialListVisitor<L, <L as MaybeMeta>::Length> {}

impl<L: MaybeMeta, V> ListVisitor<L> for V
where
    V: PartialListVisitor<L, <L as MaybeMeta>::Length>
{}

/// user-facing API to visit a list
pub fn visit<L, V>(l: &L, v: &mut V)
where
    L: MaybeMeta,
    V: ListVisitor<L>,
{
    v.visit_partial(l)
}
